/*
**  fd2lib (M68k) by Volker Barthelmann
**  rework 09/96 by Johnny Teveen <j.tevessen@line.org>
**  PPC/SVR4 version by Volker Barthelmann in 12/97
**======================================================================
**  PowerPC/WarpOS version (fd2libWOS) and complete rework by
**  Frank Wille <frank@phoenix.owl.de> in 1998
**======================================================================
**  fd2libWOS is freeware and part of the portable and retargetable
**  ANSI C compiler vbcc, copyright (c) 1995-98 by Volker Barthelmann.
**  fd2libWOS may be freely redistributed as long as no modifications
**  are made and nothing is charged for it. Non-commercial usage is
**  allowed without restrictions.
**======================================================================
**
**  v1.1  08-Jul-98 phx
**        Deleted definition of @_-symbols. vlink will create them on demand.
**        Renamed into fd2libWOS.
**        vbcc-WarpOS is ready for beta stage.
**  V1.0  16-Jun-98 phx
**        vbccppc support direct TOC access. DIRECT_TOC_ACCES = 1.
**  V0.9  02-Jun-98 phx
**        Most AmigaOS prototypes for vararg functions force the first
**        vararg into a register. A table contains those functions,
**        which has a 'correct' prototype:
**        EasyRequest,BuildEasyRequest,FWritef,FPrintf,Printf
**  V0.8  01-Jun-98 phx
**        Complete rework.
**        The new keyword '##abi' is supported, to be able to generate
**        stub routines for M68k *and* PPC functions.
**  V0.7  30-May-98 phx
**        Fixed stack frame offsets, when accessing argumentes on
**        the stack.
**  V0.5  21-Apr-98 phx
**        Wrongs offsets.
**  V0.3  10-Mar-98 phx
**        Prefix for TOC symbols is "@_" and not "@".
**  V0.2  03-Mar-98 phx
**        copied fd2libppc.c
*/

#define NDEBUG

/* only non-public (alpha) vbcc-wos releases had a zero here, */
/* it's set to 1 since beta stage */
#define DIRECT_TOC_ACCESS 1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>

#define VERSION 1
#define REVISION 1
/* AmigaOS version string. Doesn't hurt... */
const char VersTag[] = "\0$VER: fd2lib WOS 1.1 (08.07.98)";

#define MAXLINELEN      1000
#define BUFFEREDLEN     16384   /* 1024 is standard */

/* FD-file processing modes */
#define SMALLDATA        1      /* Small data access (currently no effect) */
#define VARGSLOGIC       2      /* detect ...A and ...TagList functions */
#define DEBUG            4      /* Debugging mode */

#define NUMREGS         16      /* number of M68k register in FD files */
static const char *regnames[NUMREGS] =
{
  "a0", "a1", "a2", "a3", "a4", "a5", "a6", "a7",
  "d0", "d1", "d2", "d3", "d4", "d5", "d6", "d7"
};

enum
{
  A0=0, A1, A2, A3, A4, A5, A6, A7,
  D0  , D1, D2, D3, D4, D5, D6, D7
};

/* ABIs */
#define ABI_68k         0
#define ABI_68k_NAME    "M68k"
#define ABI_PPC         1
#define ABI_PPC_NAME    "PPC"

typedef short BOOL;

#ifndef   TRUE
#  define TRUE  1
#  define FALSE 0
#endif

#ifndef   NULL
#  define NULL ((void *)0L)
#endif


static const char *varargs[] =
{
#include "vargs.h"
  NULL, NULL
};


/* these (few) OS functions have correct varargs prototypes */
static const char *std_vargs[] =
{
#include "stdvargs.h"
  NULL
};




static void ExitFailure(const char *cause,...)
{
  va_list vl;

  va_start(vl,cause);
  vfprintf(stderr,cause,vl);
  va_end(vl);
  exit(EXIT_FAILURE);
}


static void check(const char *ptr)
{
  if(!*ptr)
    ExitFailure("Unexpected EOL\n");
}


static void warnhim(int linenr, const char *format, ...)
{
  char linebuf[250];
  va_list vl;

  va_start(vl,format);
  vsprintf(linebuf, format, vl);
  va_end(vl);

  fprintf(stderr, "Warning line %d: %s\n", linenr, linebuf);
}


static FILE * OpenLVO(const char *name, const char *outdir,
                      const char *outform)
/* The LVO file is currently genereted for an M68k assembler. */
/* It will get the extension .asm, instead of .s */
{
  FILE *lvos;
  char lvoname[MAXLINELEN];

  strcpy(lvoname, outdir);

  if(name) {
    char *k = (char *)name, *p;
    int lvonamlen;

    if((p = strrchr(k, '/')) != NULL) k = p + 1;
    if((p = strrchr(k, ':')) != NULL) k = p + 1;

    strcat(lvoname, k);

    lvonamlen = strlen(lvoname);

    if((lvonamlen > 7) && !strcmp(lvoname+lvonamlen-7, "_lib.fd"))
      lvoname[lvonamlen-7] = '\0';
    else if((lvonamlen > 3) && !strcmp(lvoname+lvonamlen-3, ".fd"))
      lvoname[lvonamlen-3] = '\0';

    strcat(lvoname, "_lvo.asm");
  }
  else
    strcat(lvoname, "fd_lvo.asm");

  printf(outform, lvoname, lvoname, lvoname);

  lvos = fopen(lvoname, "w");
  fprintf(lvos,"\tsection\t.data,data\n\n");  /* Amiga68k assembler syntax */
  return(lvos);
}


static void ProcessFD(const char *name, int mode, const char *outdir,
                      const char *outform)
{
  FILE *fh, *lvos;
  FILE *out;
  int offset = 0, i, j, count, savecount;
  char function[80], tmpfuncnam[80], ff[MAXLINELEN+8], base[50];
  char line[MAXLINELEN];
  register char *p;
  char *functionp;
  int reg[NUMREGS], loops, linenr = 0, stksize, abi = ABI_68k;
  BOOL first_va_in_reg, public = -1;
  /*
  **  "-1" is the initial value for "public". It means:
  **  no statement yet.
  */

  *function = *base = *line = '\0';

  /* open FD file */
  if (name) {
    if (!(fh = fopen(name, "r")))
      ExitFailure("Could not open `%s'\n", name);
  }
  else
    fh = stdin;
  setvbuf(fh, NULL, _IOFBF, BUFFEREDLEN);             /* maybe _IOLBF? */

  /* open LVO file */
  if (!(lvos = OpenLVO(name, outdir, outform)))
    ExitFailure("Could not create lvo file\n");
  setvbuf(lvos, NULL, _IOFBF, BUFFEREDLEN);

  for(;;)
  {
    char *k;

    if (!fgets(line, MAXLINELEN-1, fh))
      ExitFailure("Unexpected EOF\n");

    linenr ++;
    if ((*line == '*') || !*line)
      continue;

    switch(*line + line[2]) {
      case ('#'+'a'):
        if (!strncmp(line, "##abi", 5)) {               /* ##abi */
          p = line+5;
          while (isspace((unsigned)*p)) p++;
          if (!strncmp(p, ABI_68k_NAME, sizeof(ABI_68k_NAME)-1)) {
            abi = ABI_68k;
            if(mode & DEBUG)
              printf("ABI set to M68k\n");
          }
          else if (!strncmp(p, ABI_PPC_NAME, sizeof(ABI_PPC_NAME)-1)) {
            abi = ABI_PPC;
            if(mode & DEBUG)
              printf("ABI set to PPC (WarpOS)\n");
          }
          else
            warnhim(linenr, "Unknown ##abi specification ignored!");
          continue;
        }
        break;

      case ('#'+'b'):
        if (!strncmp(line, "##base", 6)) {              /* ##base */
          if (*base)
            warnhim(linenr, "##base detected more than once!");
          p = line+6;
          while (isspace((unsigned)*p)) p++;
          k = base;
          while (isgraph((unsigned)*p)) *k++ = *p++;
          *k = '\0';
          if(mode & DEBUG)
            printf("Base set to `%s'\n", base);
          continue;
        }

        if (!strncmp(line, "##bias", 6)) {               /* ##bias */
          p = line+6;
          while (isspace((unsigned)*p)) p++;
          sscanf(p, "%i", &offset);
          if (mode & DEBUG)
            printf("Bias set to -%d\n", offset);
          continue;
        }
        break;

      case ('#'+'p'):
        if (!strncmp(line, "##public", 8)) {            /* ##public */
          if (public == TRUE)
            warnhim(linenr, "##public after ##public detected!");
          else
            public = TRUE;
          if (mode & DEBUG)
            printf("Turned on public mode at bias -%d\n", offset);
          continue;
        }

        if (!strncmp(line, "##private", 9)) {           /* ##private */
          if (public == FALSE)
            warnhim(linenr, "##private after ##private detected!");
          else
            public = FALSE;
          if (mode & DEBUG)
            printf("Turned off public mode at bias -%d\n", offset);
          continue;
        }
        break;

      case ('#'+'e'):
        if(!strncmp(line, "##end", 5))                  /* ##end */
          return;
        break;
    }

    if (*line == '#') {
      warnhim(linenr, "Unknown directive: `%s'!", line);
      continue;
    }

    if (public == FALSE)
    {
      offset += 6;
      continue;
    }
    if (public == -1) {
      warnhim(linenr, "Neither ##public nor ##private specified yet. "
                      "Assuming ##public.");
      public = TRUE;
    }
    functionp = function;

    /* Stub Generation Loops */
    for (loops=0; loops<=1; loops++) {
      char *p = line;
      char *k = functionp;

      while (isspace((unsigned)*p)) p++;

      if (loops == 0) {
        /* get function name */
        while ((*p != '(') && *p)
          *k++ = *p++;
        check(p);
        *k = '\0';

        /* create entry for function name in _LVOxxx file */
        fprintf(lvos, "_LVO%s\tequ\t-%d\n"
                      "\txdef\t_LVO%s\n",
                      functionp, offset, functionp);
      }
      else {
        /* skip function name */
        while ((*p != '(') && *p)
          p++;
        check(p);
      }

      if(mode & DEBUG)
        printf("function=%s, loops=%d\n", functionp, loops);

      /* Open function stub source file */
      sprintf(ff, "%s%s.s", outdir, functionp);
      printf(outform, ff, ff, ff);
      out = fopen(ff, "w");
      if(!out)
        ExitFailure("Could not create <%s>\n", functionp);
      setvbuf(out, NULL, _IOFBF, BUFFEREDLEN);

      switch (abi) {

        case ABI_68k:

          /* Write assembler headers */
#if DIRECT_TOC_ACCESS
          fprintf(out,"\t.file\t\"%s.s\"\n\t.extern _PowerPCBase\n"
                  "\t.extern %s\n\t.global _%s\n",
                  functionp,base,functionp);

#else /* as long as vbcc doesn't support direct TOC-access... */
          fprintf(out,"\t.file\t\"%s.s\"\n\t.extern _PowerPCBase\n"
                  "\t.extern @_%s\n\t.global _%s\n",
                  functionp,base,functionp);
#endif

          fprintf(out,"\t.text\n\t.align\t3\n_%s:\n",functionp);

          /* Set all registers to 'unused' */
          for(i=0; i<NUMREGS; reg[i++]=0)

          /* Skip argument names */
          while ((*p!=')') && *p) p++;
          check(p);
          p++;

          /* Search for beginning of register list */
          while ((*p!='(') && *p) p++;
          check(p);
          p++;

          /* Scan register list */
          count = savecount = 0;

          while ((*p!=')') && *p) {
            /* Check whether register description is valid */
            if ((!((*p=='a') || (*p=='A') || (*p=='d') || (*p=='D'))) ||
                 !((p[1]>='0') && (p[1]<='7')))
              ExitFailure("Bad register description '%c%c'\n",p[0],p[1]);

            /* Convert description to internal enum format */
            if ((*p=='a') || (*p=='A'))
              j = p[1] -  '0';
            else
              j = p[1] - ('0'-8);

            /* Mark register as used and save its argument counter */
            reg[j] = ++count;

            /* Increase counter if register has to be saved */
            if (!(j==A0 || j==A1 || j==A6 || j==D0 || j==D1))
              savecount ++;

            /* Search for next register */
            p += 2;
            while (isspace((unsigned)*p) && *p) p++;
            if ((*p == '/') || (*p==',')) p++;
            else
              if (*p != ')')
                ExitFailure("Parse error - ')' expected\n");
            while (isspace((unsigned)*p) && *p) p++;
            check(p);
          }

          fprintf(out,"\tmflr\tr0\n\tstw\tr0,8(r1)\n"
                      "\tstwu\tr1,-0xb0(r1)\n");

          /* clear PP_Flags, PP_Stack and PP_StackSize */
          fprintf(out,"\tli\tr11,0\n\tstw\tr11,0x28(r1)\n"
                  "\tstw\tr11,0x2c(r1)\n\tstw\tr11,0x30(r1)\n");

          /* set library base and offset */
#if DIRECT_TOC_ACCESS
          fprintf(out,"\tlwz\tr0,%s(r2)\n\tli\tr11,-%d\n"
                  "\tstw\tr0,0x20(r1)\n\tstw\tr11,0x24(r1)\n"
                  "\tstw\tr0,0x6c(r1)\n",base,offset);

#else /* as long as vbcc doesn't support direct TOC-access... */
          fprintf(out,"\tlwz\tr11,@_%s(r2)\n\tlwz\tr0,0(r11)\n\tli\tr11,-%d\n"
                  "\tstw\tr0,0x20(r1)\n\tstw\tr11,0x24(r1)\n"
                  "\tstw\tr0,0x6c(r1)\n",base,offset);
#endif

          for (i=0; i<NUMREGS; i++) { /* Registers */
            if (reg[i]) {
              if (!loops || reg[i]<count) {
                if (reg[i] <= 8)
                  fprintf(out,"\tstw\tr%d,",reg[i]+2);
                else
                  fprintf(out,"\tlwz\tr11,%d(r1)\n"
                          "\tstw\tr11,",reg[i]*4+196);
              }
              else{
                /* VarArgs */
                if (loops && first_va_in_reg) {
                  /* copy register into varargs stack */
                  fprintf(out,"\tstw\tr%d,%d(r1)\n",
                          reg[i]+2,0xc4+count*4);
                }
                fprintf(out,"\taddi\tr11,r1,%d\n\tstw\tr11,",
                        0xc4+count*4);
              }
              if (i <= 7)
                fprintf(out,"%d(r1)\n",0x54+4*i);
              else
                fprintf(out,"%d(r1)\n",0x34+4*(i-8));
            }
          } /* for i */

          /* Now place the real function call */
          fprintf(out,"\tlwz\tr3,_PowerPCBase(r2)\n\taddi\tr4,r1,0x20\n"
                  "\tlwz\tr0,-298(r3)\n\tmtlr\tr0\n\tblrl\n"
                  "\tlwz\tr3,0x34(r1)\n");
          fprintf(out,"\taddi\tr1,r1,0xb0\n\tlwz\tr0,8(r1)\n"
                  "\tmtlr\tr0\n\tblr\n"
                  "\t.type\t_%s,@function\n"
                  "\t.size\t_%s,$-_%s\n\n",
                  functionp,functionp,functionp);
          fclose(out);
          break;

        case ABI_PPC:

          /* Write assembler headers */
          fprintf(out,"\t.file\t\"%s.s\"\n\t.text\n"
#if DIRECT_TOC_ACCESS
                  "\t.extern %s\n\n"
#else
                  "\t.extern @_%s\n\n"
#endif
                  "\t.align\t3\n\t.globl\t_%s\n_%s:\n",
                  functionp,base,functionp,functionp);

          /* Search for beginning of argument list */
          while (*p!='(' && *p) p++;
          check(p);
          p++;

          /* determine the number of arguments */
          count = 0;
          for (;;) {
            while (*p!=')' && isspace((unsigned)*p) && *p)
              ++p;
            check(p);
            if (*p == ')')
              break;
            count++;
            while (*p!=')' && *p!=',' && *p)
              ++p;
            check(p);
            if (*p++ == ')')
              break;
          }
          if (mode & DEBUG)
            printf("PPC function %s has %d arguments.\n",functionp,count);
          if (count > 26)
            ExitFailure("PPC function %s exceeds the current limitation"
                        "of 26 arguments by %d!",functionp,count-26);

          /* init stack frame */
          stksize = (count<8) ? 32 : ((56+(count-7)*8+15)&~15);
          fprintf(out,"\tmflr\tr0\n\tstw\tr0,8(r1)\n"
                      "\tstwu\tr1,-%d(r1)\n",stksize);

          if (count > 7) {
            /* extra arguments must be passed on the stack */
            if (count == 8) {
              /* special case: move 8th argument into stack frame */
              if (loops)  /* varargs? */
                fprintf(out,"\taddi\tr10,r1,%d\n",stksize+20+count*4);
              fprintf(out,"\tstw\tr10,56(r1)\n");
            }
            else {
              int firstreg = 32-(count-7);

              fprintf(out,"\tstmw\tr%d,%d(r1)\n"
                      "\tmr\tr%d,r10\n"
                      "\tlmw\tr%d,%d(r1)\n",
                      firstreg,56+(count-7)*4,firstreg,
                      firstreg+1,stksize+56);
              if (loops)  /* varargs? */
                fprintf(out,"\taddi\tr31,r1,%d\n",stksize+20+count*4);
              fprintf(out,"\tstmw\tr%d,56(r1)\n",firstreg);
            }
          }
          else if (loops) {  /* pass varargs pointer in a register? */
            fprintf(out,"\taddi\tr%d,r1,%d\n",count+3,stksize+20+count*4);
            /* ATTENTION: count will be modified and must not be */
            /*            used for calculations after this point! */
            --count;
          }

          /* shift all arguments into their following register */
          for (i=(count<8)?count:7; i>0; i--)
            fprintf(out,"\tmr\tr%d,r%d\n",3+i,2+i);

          /* load library base and LVO, then call LVO via LR */
          fprintf(out,
#if DIRECT_TOC_ACCESS
                  "\tlwz\tr3,%s(r2)\n"
#else
                  "\tlwz\tr3,@_%s(r2)\n\tlwz\tr3,0(r3)\n"
#endif
                  "\tlwz\tr0,-%d(r3)\n"
                  "\tmtlr\tr0\n"
                  "\tblrl\n",base,offset-2);

          /* cleanup stack frame and return */
          if (count > 8)
            fprintf(out,"\tlmw\tr%d,%d(r1)\n",  /* restore saved regs */
                    32-(count-7),56+(count-7)*4);
          fprintf(out,"\taddi\tr1,r1,%d\n"
                  "\tlwz\tr0,8(r1)\n\tmtlr\tr0\n\tblr\n"
                  "\t.type\t_%s,@function\n"
                  "\t.size\t_%s,$-_%s\n\n",
                  stksize,functionp,functionp,functionp);

          fclose(out);
          break;

      } /* switch */

      if (loops > 0) break;

      p = (char *)*varargs;
      loops = 3;
      if (mode & DEBUG)
        printf("Searching function `%s' in vargs table...\n", functionp);

      if (p != NULL) {
        if (mode & VARGSLOGIC) {
          int fnlen = strlen(functionp);

          if ((fnlen > 7) && (!strcmp(functionp+fnlen-7, "TagList"))) {
            /* xxxTagList function found. Make xxxTags of it */
            strncpy(tmpfuncnam, functionp, fnlen-4);
            strcpy(tmpfuncnam+fnlen-4, "s");
            functionp = tmpfuncnam;
            loops = 0;
          }

          else if ((fnlen > 1) && ((functionp[fnlen-1] == 'A') &&
                   (functionp[fnlen-2] >= 'a') && 
                   (functionp[fnlen-2] <= 'z'))) {
            /*  Not that smart recognition... But you probably
            **  don't want to have a function CreateDA() varargs,
            **  want you?
            **
            **  Recognized are functions that end with 'A'and that
            **  have a lowercase letter before that.
            */
            strcpy(tmpfuncnam, functionp);
            tmpfuncnam[fnlen-1] = '\0';
            functionp = tmpfuncnam;
            loops = 0;
          }

          if (!loops && (mode&DEBUG))
            puts("Found via internal logic!");
        }

        if (loops) {
          for(i=0; p!=NULL; i+=2) {
            if (!strcmp(p, functionp)) {
              char **stdva = (char **)std_vargs;

              if (mode & DEBUG)
                puts("Found!");
              functionp = (char *)varargs[i-1];
              loops = 0;

              /* The AmigaOS clib/#?.h header files unfortunately */
              /* declare most vararg functions with the first vararg */
              /* appearing in the argument list. */
              /* vbcc's PowerOpen code generator will pass the first */
              /* vararg in a register then. :(  */
              first_va_in_reg = TRUE;

              /* uses correct prototype? */
              while (*stdva) {
                if (!strcmp(*stdva,functionp)) {
                  if (mode & DEBUG)
                    printf("%s() assumed to use correct prototype.\n",
                           functionp);
                  first_va_in_reg = FALSE;
                  break;
                }
                stdva++;
              }
              break;
            }
            p = (char *)varargs[i];
          }
        }
      }
    }

    offset += 6;
  } /* for(;;) */

  if(name)
    fclose(fh);
  fclose(lvos);
}


static void fillpath(char *dirpath)
/*  Append '/' to path if needed */
{
  int sl = strlen(dirpath);

  switch(dirpath[sl-1]) {
    case ':':
    case '/':
      break;
    default:
      strcpy(dirpath+sl, "/");
      break;
  }
}


static void Usage(const char *myname)
/*  Show program usage */
{
  printf("fd2lib WarpOS V%d.%d\n"
         "(c)1996-98 by Frank Wille, Volker Barthelmann, Johnny Teveen\n"
         "\n"
         "  -- Caution: Needs ~5000 byte stack! --\n"
         "\n"
         "Usage: %s [-sd] [-nv] [-o <dir>] [-of <format>] "
         "[-d] [-?|--help] [files/pattern]\n"
         "\nOptions:\n"
         "-sd    Use small data model (currently not supported)\n"
         "-nv    No varargs logic - ...A and ...TagList will not be detected\n"
         "-o     Specify directory to store source files in\n"
         "-of    C printf style output format to generate compiling\n"
         "       script. Three `%%s' are replaced with output file name\n"
         "-d     Turn on debugging/verbose mode\n"
         "-?     Show help/version and quit\n"
         "files  FD files to convert, defaults to stdin\n"
         "\n"
         "Commandline is parsed left-to-right. Specifying\n"
         "\"alib_lib.fd -sd blib_lib.fd\" will result in alib\n"
         "getting large data model.\n",VERSION,REVISION,myname);
  exit(0);
}


int main(int argc, char *argv[])
{
  int erg       = 0  /* EXIT_FAILURE */;
  int mode      = VARGSLOGIC;
  int filesdone = 0;

  char outdir[80] = "", outform[250] = "";

  if(argc > 1) {
    int i;

    for(i=1; i<argc; i++) {
      if(argv[i][0] == '-') {
        /* Parse option */

        if( !strcmp(argv[i], "-sd")) mode |=  SMALLDATA;
        else if( !strcmp(argv[i], "-nv")) mode &= ~VARGSLOGIC;
        else if( !strcmp(argv[i], "-o" )) {
          if(i < (argc-1)) {
            i ++;
            if(strlen(argv[i]) < sizeof(outdir)) {
              strcpy  (outdir, argv[i]);
              fillpath(outdir);
            }
            else {
              fprintf(stderr, "Path too long. Maximum is %lu characters. Ignored.\n", (unsigned long)sizeof(outdir));
            }
          }
          else {
            fputs("No path specified after `-o'!\n", stderr);
          }
        }
        else if( !strcmp(argv[i], "-of")) {
          if(i < (argc-1)) {
            i ++;
            if(strlen(argv[i]) < sizeof(outform)) {
              strcpy(outform, argv[i]);
              strcat(outform, "\n"   );
            }
            else {
              fprintf(stderr, "Format too long. Maximum is %lu characters. Ignored.\n", (unsigned long)sizeof(outform));
            }
          }
          else {
            fputs("No format specified after `-of'!\n", stderr);
          }
        }
        else if( !strcmp(argv[i], "-d" )) mode |=  DEBUG;
        else if((!strcmp(argv[i], "-?" )) ||
                (!strcmp(argv[i], "--help"))) Usage(*argv);
        else {
          fprintf(stderr, "Unknown option `%s'\n\n", argv[i]);
          Usage(*argv);
        }
      }
      else {
        /* Process file */
        if(argv[i][0] == '?')
          Usage(*argv);
        else {
          ProcessFD(argv[i], mode, outdir, outform);
          filesdone ++;
        }
      }
    }
  }

  if(!filesdone)
    ProcessFD(NULL, mode, outdir, outform);

  return(erg);
}
